# -*- coding: utf-8 -*-
# Copyright (C) 2005 Lateef Alabi-Oki
#
# This file is part of Scribes.
#
# Scribes is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Scribes is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Scribes; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""
This module defines functions that manipulate cursor activities such as changing
the changing, showing and hiding the cursor. The modules also contains the
function that displays the cursor position on the status bar.

@author: Lateef Alabi-Oki
@organiation: The Scribes Project
@copyright: Copyright © 2005 Lateef Alabi-Oki
@license: GNU GPLv2 or Later
@contact: mystilleef@gmail.com

"""

from re import search, UNICODE

from pygtk import require
require("2.0")
from gtk import TEXT_WINDOW_TEXT
from gtk.gdk import Cursor, WATCH, XTERM
from internationalization import msg0329


def cursor_position(editor):
	"""
	Ascertains the position of the cursor at all times

	@param editor: Reference to an editor instance
	@type editor: An editor object

	@return: The cursor's position
	@rtype: A tuple object containing the cursor line and index.

	"""

	# We need three components to ascertain the line number the cursor is
	# on: the cursor's mark; the cursor's iter; and the line the cursor's
	# iter is on. The cursor's mark is a default and preset value defined by
	# the text buffer, thankfully. So we retrieve the mark by simply
	# querying the text buffer for it. We get the cursor's iter from the
	# cursor's mark, and in turn get the cursor's line from the cursor's
	# iter. This gives us the line number the cursor is on.

	# The line number the cursor is on.
	cursor_mark = editor.text_buffer.get_insert()
	cursor_iter = editor.text_buffer.get_iter_at_mark(cursor_mark)
	cursor_index = cursor_iter.get_line_index()
	cursor_line = cursor_iter.get_line() + 1

	# To ascertain the column the cursor is on, we simply count the number
	# of characters before the cursor on the cursor line and add 1 to the
	# number. If the character is a tab character, we ask the text view
	# widget to give us its size, and use it to calculate the count
	# increase.
	#
	# We need three key ingredients for the magic to happen: the iterator
	# at the beginning of the line; the iterator at the cursor position; and
	# the characters between both iterators. Afterwhich, we loop through the
	# characters and count taking into consideration tab characters as
	# discussed above. I'm not sure if this is the most efficient way to do
	# it, but so far, it works correctly.

	# The column the cursor is on.
	count = 0
	start_iter = editor.text_buffer.get_iter_at_line(cursor_iter.get_line())
	line_text = editor.text_buffer.get_text(start_iter, cursor_iter)
	tab_size = editor.text_view.get_tabs_width()

	for tabs in line_text:

		if tabs == "\t":

			count += (tab_size - (count % tab_size))

		else:

			count += 1

	cursor_column = count + 1

	# Display the cursor's position on the status bar.
	string = msg0329 % (cursor_line, cursor_column)
	editor.status_two.pop(editor.status_two_id)

	editor.status_two_id = editor.status_two.get_context_id(string)
	editor.status_two.push(editor.status_two_id, string)

	return cursor_line, cursor_index


def show_cursor(editor, pattern=None):
	"""
	Show or hide the busy cursor in the editor's buffer area.

	The busy cursor should be shown anytime a long operation may occur.

	@param editor: Reference to the editor instance
	@type editor: An editor object

	@param pattern: Kind of cursor to show. Only WATCH and XTERM are supported.
	@type pattern. gtk.gdk.Cursor type

	"""

	if pattern == "busy":

		pattern = WATCH

		# A flag that indicates the cursor is busy and should not be changed.
		editor.busy_cursor = True

	else:

		pattern = XTERM

		# A flag that indicates the cursor is not busy and can be hidden or
		# changed.
		editor.busy_cursor = False

		try:

			editor.hidden_cursor = False

		except AttributeError:

			editor.hidden_cursor = False

	try:

		textview_win = editor.text_view.get_window(TEXT_WINDOW_TEXT)
		textview_win.set_cursor(Cursor(pattern))

	except:

		pass

	return


def get_cursor_iter(editor):
	"""
	Get the cursor's iterator in the editor's buffer.

	@param editor: Reference to the editor object
	@type editor: An editor object

	@return: The cursor's iterator.
	@rtype: A gtk.Iter.

	"""

	cursor_mark = editor.text_buffer.get_insert()
	cursor_iter = editor.text_buffer.get_iter_at_mark(cursor_mark)

	return cursor_iter


def get_current_line(editor):
	"""
	Get the line number the cursor is on in the editor's buffer.

	@param editor: Reference to the editor object.
	@type editor: An editor object.

	@return: The line the cursor is on.
	@rtype: An integer object.

	"""

	cursor_iter = get_cursor_iter(editor)
	current_line = cursor_iter.get_line()

	return current_line


def get_string_to_cursor(editor):
	"""
	Get strings from beginning of a word to cursor

	@param editor: Reference to the editor object
	@type editor: An editor object

	@return string: Characters from beginning of a word to cursor
	@rtype: A string

	"""

	cursor_iter = get_cursor_iter(editor)

	# If the cursor is at the beginning of a line return nothing.
	if cursor_iter.starts_line():

		return

	# If the cursor is inside a word do nothing. This gets rid of some
	# annoying automatic completion behavior.
	if cursor_iter.get_char().isalnum() or cursor_iter.get_char() == "_":

		return

	# Get the iterator at the beginning of the cursor line.
	start_iter = editor.text_buffer.get_iter_at_line(get_current_line(editor))

	# Get text from beginning of the cursor line to the cursor point.
	text = editor.text_buffer.get_text(start_iter, cursor_iter,
									 include_hidden_chars=False)

	string = text.decode("utf8")

	# Apply regular expression on text to get the last string to the cursor
	# point.
	match = search(editor.cursor_pattern, string, UNICODE)

	# If a match is found and the match is at least 3 characters long, then
	# return the string for automatic completion anaylsis. Otherwise, return
	# nothing.
	if match:

		if len(match.group()) >= 3:

			return match.group()

	return


def get_cursor_window_coordinates(editor):
	"""
	Get the window coordinates of the cursor in the text editor' buffer.

	@param editor: Reference to the editor object.
	@type editor: An editor object

	@return: The position of the cursor in the text editor's buffer
	@rtype: A tuple representing the x and y coordinates of the cursor's
			position in the text editor' buffer.

	"""

	view = editor.text_view

	# Get the cursor's iterator.
	cursor_iter = get_cursor_iter(editor)

	# Get the cursor's buffer coordinates.
	rectangle = view.get_iter_location(cursor_iter)

	# Get the cursor's window coordinates.
	position = view.buffer_to_window_coords(TEXT_WINDOW_TEXT, rectangle.x,
											rectangle.y)
	cursor_x = position[0]
	cursor_y = position[1]

	return cursor_x, cursor_y


def get_cursor_size(editor):
	"""
	Get the cursor's size.

	@param editor: Reference to the editor object.
	@type editor: An editor object

	"""

	view = editor.text_view

	# Get the cursor's iterator.
	cursor_iter = get_cursor_iter(editor)

	# Get the cursor's size via its buffer coordinates.
	rectangle = view.get_iter_location(cursor_iter)
	cursor_width = rectangle.width
	cursor_height = rectangle.height

	return cursor_width, cursor_height


def move_view_to_cursor_position(editor):
	"""
	Move the text editor's view to the cursor position.

	@param editor: Reference to the editor instance.
	@type editor: An editor object.

	"""

	cursor_mark = editor.text_buffer.get_insert()
	editor.text_view.scroll_to_mark(cursor_mark, 0.3,
								 use_align=False, xalign=0.5,
								 yalign=0.5)

	return


def get_characters_to_cursor(editor):
	"""
	Get characters after a space to the cursor position.

	@param editor: Reference to the editor instance.
	@type editor: An editor object.

	@return: Return a string if and its beginning and ending position in the
		text editor's buffer.
	@rtype: A tuple object.

	"""

	# Get the cursor's position.
	cursor_iter = get_cursor_iter(editor)
	start_iter = cursor_iter.copy()
	start_iter.backward_char()

	# Get the characters before the cursor but after a space, tab or newline.
	while not start_iter.get_char() in (" ", "\t") and \
	 	not start_iter.is_start():

		if start_iter.starts_line():

			if start_iter.get_char() in ("\n"):

				start_iter.forward_line()

			elif start_iter.get_char in (" ", "\t"):

				start_iter.forward_char()

			break

		start_iter.backward_char()


	if start_iter.starts_line():

		if start_iter.get_char() in ("\n"):

			start_iter.forward_line()

		elif start_iter.get_char() in (" ", "\t"):

			start_iter.forward_char()

	else:

		start_iter.forward_char()

	text = editor.text_buffer.get_text(start_iter, cursor_iter)

	return text, start_iter, cursor_iter

