# -*- coding: utf-8 -*-
# Copyright (C) 2005 Lateef Alabi-Oki
#
# This file is part of Scribes.
#
# Scribes is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Scribes is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Scribes; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


"""
File loading module for Scribes

This module describes a class that performs operations necessary to load a
file. This module also exposes a launch_new_file_timeout() callback. The
callback is activated only when more than one file is being loaded. When only
one file is being loaded, the callback is ignored.

@note:	The text editor follows a document centric and spatial paradigm. This
		means each file is loaded in its own editor window. Conceptually, think
		of the editor windows as a document, rather than an application.

"""

from gc import collect
from os import path, environ, R_OK, W_OK, F_OK, access
from locale import getpreferredencoding

from pygtk import require
require("2.0")
from gobject import timeout_add
from gtk import main as gmain, quit_add
from gnomevfs import get_file_info, get_mime_type
from gnomevfs import get_local_path_from_uri
from gnomevfs.async import open as gvfsopen

from syntax import syntax_highlight
from error import error_dialog
from feedback import update_status_info
from timeout import activate_progressbar_timeout
from position import get_metadata_info
from internationalization import msg0025, msg0030, msg0031, msg0032, msg0033
from internationalization import msg0034, msg0035, msg0036, msg0037, msg0153
from internationalization import msg0191, msg0192
from cursor import show_cursor
from bookmark import bookmark_line
from readonly import activate_readonly_mode
import main


class FileLoader(object):
	"""
	Defines attributes and methods necessary to load a file.

	The sole function of this class is to load a text file into the editor's
	buffer. The	class exposes methods that performs numerous checks to ensure
	only text files are loaded into the editor's buffer. Hence, only text files
	that the user has permission to view can be loaded into the editor's buffer.

	"""


	def __init__(self, editor):
		"""
		Initialize a FileLoader instance.

		@param self: Reference to a FileLoader instance
		@type self: a FileLoader object

		@param editor: 	Reference to the editor instance
		@type editor: An editor object

		@param encoding: The encoding that should be used to open the editor.
		@type encoding: A string

		"""

		self.editor = editor
		self.encoding = self.editor.filechooser_encoding
		show_cursor(self.editor, "busy")
		self.iter = None
		self.message_id = None
		self.blocked_signals = False
		self.mime_list = ["image/svg+xml", "application/x-shellscript",
						  "application/rtf", "application/x-glade",
						  "application/x-php", "application/octet-stream",
						  "application/x-desktop", "application/x-trash",
						  "application/x-csh", "audio/x-mpegurl",
						  "application/x-perl"]


	def check_load_flag(self, filename):
		"""
		Determines whether or not the file should be loaded in a new window

		Each document is loaded in its own window. So if a window already has
		a file loaded, a new editor window is created and the new file is loaded
		into the newly created window.

		@param self: Reference to a FileLoader instance.
		@type self: A FileLoader object.

		@param filename: The name of the file to be loaded.
		@type filename: A string.

		"""

		# If the buffer is not empty, or the buffer has been modified, load the
		# file in a new editor instance.
		if self.editor.text_buffer.get_char_count() or \
			self.editor.text_buffer.get_modified() or self.editor.loading_flag \
			or self.editor.filename !=  msg0025:

			show_cursor(self.editor)
			quit_add(0, gmain)
			timeout_add(2, self.launch_new_file_timeout, filename)

		else:

			self.editor.loading_flag = True
			self._redefine_filename(filename)

		return


	def _redefine_filename(self, filename):
		"""
		Define the file's URI, filename and statusbar name

		The file's URI is used by GNOMEVFS to load files. The filename is the
		full path to the file. And the statusbar name is the full path to the
		file as it will appear in the statusbar.

		@param self: Reference to a FileLoader instance.
		@type self: A FileLoader object.

		@param filename: The name of the file to be loaded.
		@type filename: A string.

		"""

		self.editor.file_uri = filename.strip()
		temp = get_local_path_from_uri(self.editor.file_uri)
		self.editor.filename = temp.strip()

		if self.editor.filename.startswith(environ['HOME']):

			filename = self.editor.filename.replace(environ['HOME'], "~")
			self.editor.prettyname = filename.strip()

		else:

			self.editor.prettyname = self.editor.filename.strip()

		self._check_metadata_database(self.editor.filename)

		return


	def _check_metadata_database(self, filename):
		"""
		Check the editor's metadata database for an entry of the file.

		If a file already exists in the database, skip file checks to speed up
		loading.

		@param self: Reference to a FileLoader instance.
		@type self: A FileLoader object.

		@param filename: The name of the file to be loaded.
		@type filename: A string.

		"""

		# Recall the window position, buffer state and encoding of the file if
		# it has an entry in Scribes' database.
		result = get_metadata_info(self.editor)

		if result:

			if not access(filename, R_OK):

				# Show an error dialog if the user does not have permission to
				# view the file.
				self._show_error_dialog(msg0035)

			elif not access(filename, W_OK) or self.editor.readonly:

				# Load the file in readonly mode.
				self._prepare_editor_for_loading(readonly=True)

			else:

				# Prepare to load the file.
				self._prepare_editor_for_loading()

		else:

			self._check_file_validity(filename)

		return


	def _check_file_validity(self, filename):
		"""
		Check file permissions, type and existence.

		@param self: Reference to a FileLoader instance.
		@type self: A FileLoader object.

		@param filename: The name of the file to be loaded.
		@type filename: A string.

		@note: See internationalization.py for the value of msgXXXX

		"""

		# Check the mimetype of the file. Only text and SVG files will be
		# loaded.
		mimetype = get_mime_type(self.editor.file_uri)

		if not mimetype:

			# Show an error dialog if the MIME type of the file cannot be
			# determined.
			message = msg0191 % mimetype
			self._show_error_dialog(message)

		elif not access(filename, F_OK):

			# Show the error dialog if the file does not exist.
			self._show_error_dialog(msg0037)

		elif not path.isfile(filename):

			# Show an error dialog if the file is not a text file.
			self._show_error_dialog(msg0036)

		elif not access(filename, R_OK):

			# Show an error dialog if the user does not have permission to view
			# the file.
			self._show_error_dialog(msg0035)

		elif not mimetype.startswith("text/") \
			and mimetype not in self.mime_list:

			# Show an error dialog if the MIME type of the file is not a text,
			# SVG, shellscript or rtf one.
			message = msg0192 % mimetype
			self._show_error_dialog(message)

		elif not access(filename, W_OK) or self.editor.readonly:

			# Load the file in readonly mode.
			self._prepare_editor_for_loading(readonly=True)

		else:

			# Prepare to load the file.
			self._prepare_editor_for_loading()

		return


	def _prepare_editor_for_loading(self, readonly=False):
		"""
		Set up the editor window title and status bar.

		@param self: Reference to a FileLoader instance.
		@type self: A FileLoader object.

		@param readonly: 	whether or not file is readonly
		@type readonly: 	boolean

		@note: See internationalization.py for the value of msgXXXX

		"""

		if readonly:

			activate_readonly_mode(self.editor)

		else:
			
			title = path.basename(self.editor.filename)
			self.editor.set_title(title)

		self._load_file()

		return


	def _load_file(self):
		""""
		Load the file into the editor's buffer.

		@param self: Reference to a FileLoader instance.
		@type self: A FileLoader object.

		@note: Loading a file into the buffer is an asynchronous operation.
		@note: See internationalization.py for the value of msgXXXX

		"""

		self._get_file_information()

		# Feedback to statusbar area indicating that a file is being loaded.
		update_status_info(self.editor, msg0032, "run", False)

		# Block certain buffer signals to prevent flag corruption.
		self._block_buffer_signals()

		# If the file has an entry in Scribes' database, the encoding may have
		# changed. Therefore, retrieve the new encoding.
		self.encoding =	self._recheck_file_encoding()

		# Hide the cursor.
		self._set_cursor_visibility(False)

		try:

			# Open the file for loading. This operation is asynchronous.
			gvfsopen(self.editor.file_uri, self.open_cb)

		except:

			self._show_error_dialog(msg0033)

		else:

			# Show the progress bar for files that take more than a second to
			# load.
			self.editor.gtimer = timeout_add(1000, activate_progressbar_timeout,
									 self.editor)

			# Make toolbar buttons insensitive while loading the file.
			self._set_toolbar_buttons_sensitivity(False)

		return


	def open_cb(self, handle, exec_type):
		"""
		Handles callback when a file is being open for reading

		@param self: reference to a FileLoader instance
		@type self: a FileLoader object
		@param handle: file handle
		@type handle: ?
		@param exec_type: ?
		@type exec_type: ?

		"""

		# Clear the text buffer and begin reading the file to be loaded.
		self.editor.text_buffer.set_text("")
		self.iter = self.editor.text_buffer.get_start_iter()
		self._set_buffer_editability(False)
		handle.read(self.editor.opt_size, self.read_cb)

		return


	def read_cb(self, handle, textbuffer, exc_type, bytes_requested):
		"""
		Handles callback when a file is being opened for reading

		@param self: reference to a FileLoader instance
		@type self: a FileLoader object
		@param handle: file handle
		@type handle: ?
		@param buffer: ?
		@type buffer: ?
		@param exc_type: ?
		@type exc_type: ?
		@param bytes_requested: ?
		@type bytes_requested: ?

		"""

		if textbuffer != "":

			# Progress bar information
			if self.editor.progress_flag:

				self.editor.opt_count = self.editor.opt_count + \
										self.editor.opt_size

			unicode_txt = self._decode_text_to_unicode(textbuffer)
			output = self._encode_text_to_utf8(unicode_txt)
			self._insert_text_to_buffer(output, handle)

		else:

			self.editor.progress_flag = False
			message = msg0030 % self.editor.prettyname
			update_status_info(self.editor, message, "open")
			self._determine_cursor_placement()
			self._determine_bookmark_placement()
			syntax_highlight(self.editor, self.editor.file_uri)
			self._set_toolbar_buttons_sensitivity(True)
			self._set_cursor_visibility(True)
			self._set_buffer_editability(True)
			self.editor.text_buffer.set_modified(False)
			self._unblock_buffer_signals()
			self._set_finalization_flags()
			self.editor.encoding = self.encoding
			show_cursor(self.editor)
			handle.close(self.close_cb)

		return


	def close_cb(self, *args):
		"""
		Handles callback to close an open file

		@param self:	fileloader object

		"""

		del self
		collect()

		return


	def _error_recovery(self):
		"""
		Reset the editor to a neutral state after an error occurs.

		This function is called when a error occurs while a file can not be
		loaded for several reasons. The function returns the editor window to a
		neutral state as if a new window has just been launched.

		@param self:	fileloader object

		"""

		if self._blocked_buffer_signals():

			self._unblock_buffer_signals()

		self.editor.initialize_flags()
		self._set_cursor_visibility(True)
		self._set_buffer_editability(True)
		self.editor.text_buffer.set_text("")
		self._set_toolbar_buttons_sensitivity(True)
		self.editor.text_buffer.set_modified(False)
		self.editor.set_title(msg0025)
		self.editor.filename = msg0025
		update_status_info(self.editor, "", "")

		return


	def _set_toolbar_buttons_sensitivity(self, value=True):
		"""
		Disable or enable some toolbar buttons.

		To prevent users from corrupting the buffer when loading a file, some
		toolbar buttons are disabled. The replace toolbar button, in particular,
		is disabled in readonly mode. As users should not be able to perform
		such text operations in readonly mode.

		@param self: Reference to the FileLoader object.
		@type self: A FileLoader object.

		@param value: True to enable toolbar buttons. False to disable them.
		@type value: A boolean object.

		"""

		self.editor.actiongroup.get_action(u"saveas_file").set_sensitive(value)
		self.editor.actiongroup.get_action(u"replace_text").set_sensitive(value)
		self.editor.actiongroup.get_action(u"print_file").set_sensitive(value)
		self.editor.actiongroup.get_action(u"goto_line").set_sensitive(value)
		self.editor.actiongroup.get_action(u"search_text").set_sensitive(value)
		self.editor.actiongroup.get_action(u"preferences").set_sensitive(value)

		# Always disable the replace toolbar button in readonly mode.
		if self.editor.readonly:

			self.editor.actiongroup.get_action(u"replace_text").set_sensitive(False)

		return


	def _recheck_file_encoding(self):
		"""
		Check the current encoding of a file.

		The encoding of a file is determined by three factors, namely: the user;
		the editor and; the editor's database. The user has control over which
		encoding to use to open a file. In such instances, the user's encoding
		of choice overrides that of editor's or that entered in the editor's
		database.

		When a file is saved in a particular encoding, the encoding is
		associated with the file and stored in the editor's database. So that
		when next the file is loaded, the encoding stored in the database is
		used. If the user chooses no encoding, or no encoding is stored in the
		editor's database, then the editor's default encoding, UTF-8 is used
		when loading the file.

		@param self: Reference to the FileLoader object.
		@type self: A FileLoader object.

		"""

		if self.encoding:

			self.editor.encoding = self.encoding

		else:

			self.encoding = self.editor.encoding

		return self.encoding


	def _block_buffer_signals(self):
		"""
		Block key buffer signals while loading a file.

		Block the "changed" and "modified changed" text buffer signals. These
		signals should be blocked to prevent the callbacks associated with
		them from being activated when loading a file.

		@param self: Reference to the FileLoader object.
		@type self: A FileLoader object.

		"""

		self.editor.text_buffer.handler_block(self.editor.tb_mod_id)
		self.editor.text_buffer.handler_block(self.editor.tb_chg_id)
		self.editor.text_buffer.handler_block(self.editor.tb_ins_id)
		self.editor.text_buffer.begin_not_undoable_action()

		# The flag is true if buffer signals have been blocked.
		self.blocked_signals = True

		return


	def _unblock_buffer_signals(self):
		"""
		Unblock key buffer signals after a file has been loaded.

		@param self: Reference to the FileLoader object.
		@type self: A FileLoader object.

		"""

		self.editor.text_buffer.handler_unblock(self.editor.tb_mod_id)
		self.editor.text_buffer.handler_unblock(self.editor.tb_chg_id)
		self.editor.text_buffer.handler_unblock(self.editor.tb_ins_id)
		self.editor.text_buffer.end_not_undoable_action()

		# The flag is true if buffer signals have been blocked.
		self.blocked_signals = False

		return


	def _blocked_buffer_signals(self):
		"""
		Check if buffer signals are blocked.

		@param self: Reference to the FileLoader object.
		@type self: A FileLoader object.

		@return: Return True if buffer signals are blocked and false otherwise.
		@rtype: A boolean object.

		"""

		if self.blocked_signals:

			value = True

		else:

			value = False

		return value


	def _set_cursor_visibility(self, value=True):
		"""
		Show or hide the cursor.

		The cursor is hidden when the file is being loaded and shown when the
		it has finished loading. The cursor is always hidden in readonly mode.

		@param self: Reference to the FileLoader object.
		@type self: A FileLoader object.

		@param value: True to show the cursor false otherwise.
		@type value: A boolean object.

		"""

		if self.editor.readonly:

			value = False

		self.editor.text_view.set_cursor_visible(value)
		self.editor.text_view.set_highlight_current_line(value)

		return


	def _determine_cursor_placement(self):
		"""
		Determine where to place the cursor when a file is loaded.

		The editor stores the last cursor position in a database so that
		when a file is loaded the cursor is placed at the last position in the
		file.

		However, it is possible that another software program alters the state
		of the file. Thus making the information in the editor's database
		obsolete. In such a case, the editor tries to determine where to place
		the cursor.

		If the changes are too drastic, the editor places the cursor at the
		begining of the file. If the changes are minor, the editor places the
		cursor at an approximate position close the the last position the cursor
		was last seen.

		@param self: Reference to the FileLoader object.
		@type self: A FileLoader object.

		"""

		start_iter = self.editor.text_buffer.get_start_iter()

		# Place the cursor at the begining of the file if the file has no entry
		# in the database.
		if self.editor.position_flag:

			# If the number of lines in the text editor's buffer is less than
			# that stored in the database, then place the cursor at the begining
			# of the file.
			number_of_lines = self.editor.text_buffer.get_line_count()
			last_line = self.editor.pos[0][0]

			if last_line > number_of_lines:

				self.editor.text_buffer.place_cursor(start_iter)
				self.editor.text_view.queue_draw()

				return

			# If the byte index on the line is less than that stored in the
			# database, then place the cursor at the position of smaller byte
			# index.
			line_iter = \
			self.editor.text_buffer.get_iter_at_line(last_line-1)
			number_of_bytes_in_line = line_iter.get_bytes_in_line()
			last_line_index = self.editor.pos[0][1]

			if last_line_index > number_of_bytes_in_line:

				line_iter.set_line_index(number_of_bytes_in_line)

			else:

				line_iter.set_line_index(last_line_index)

			self.editor.text_buffer.place_cursor(line_iter)
			cursor_mark = self.editor.text_buffer.get_insert()
			self.editor.text_view.scroll_to_mark(cursor_mark, 0.3,
								 use_align=False, xalign=0.5,
								 yalign=0.5)

		else:

			self.editor.text_buffer.place_cursor(start_iter)

		self.editor.text_view.queue_draw()

		return


	def _determine_bookmark_placement(self):
		"""
		Check whether or not the document should be bookmarked and place
		the bookmark in the right places if it should.

		@param self: Reference to the FileLoader object.
		@type self: A FileLoader object.

		"""

		if not self.editor.readonly and self.editor.bookmark_flag:

			for lines in self.editor.bookmarked_lines:

				bookmark_line(self.editor, lines)

		return


	def _set_buffer_editability(self, value=True):
		"""
		Set or Unset whether or not the buffer should be editable.

		The buffer should not be editable in readonly mode.

		@param self: Reference to the FileLoader object.
		@type self: A FileLoader object.

		@param value: True if the buffer should be editable, False otherwise.
		@type value: A boolean object.

		"""

		if self.editor.readonly:

			value = False

		self.editor.text_view.set_editable(value)

		return


	def _decode_text_to_unicode(self, text):
		"""
		Decode text in files to Unicode.

		Unicode is the universal encoding understood by Python and GNOME.
		Strings must first be decoded to Unicode before they can be encoded to
		any other character set.

		@param self: Reference to the FileLoader object.
		@type self: A FileLoader object.

		@param text: Text retrieved from a file.
		@type text: An encoded text.

		@return: A unicode string.
		@rtype: A unicode string object.

		"""

		try:

			txt = text.decode(self.encoding, "replace")

		except UnicodeDecodeError:

			self.editor.progress_flag = False
			self._show_error_dialog(msg0031)
			txt = None

		return txt


	def _encode_text_to_utf8(self, text):
		"""
		Encode unicode text to UTF-8.

		gtk.TextBuffer only accepts UTF-8 encoded characters, thank goodness.

		@param self: Reference to the FileLoader object.
		@type self: A FileLoader object.

		@param text: Unicode text.
		@type text: A unicode string.

		@return: An encoded string.
		@rtype: A encoded string object.

		"""

		try:

			txt = text.encode("utf8")

		except UnicodeEncodeError:

			self.editor.progress_flag = False
			self._show_error_dialog(msg0153)
			txt = None

		except AttributeError:

			txt = None

		return txt


	def _insert_text_to_buffer(self, text, handle):
		"""
		Inserted UTF-8 encoded character into the text editor's buffer.

		@param self: Reference to the FileLoader object.
		@type self: A FileLoader object.

		@param text: UTF-8 encoded text.
		@type text: A UTF-8 encoded string object.

		@param handle: A reference to the file object to load into the buffer.
		@type handle: A gnomevfs handle.

		"""

		try:

			self.editor.text_buffer.insert(self.iter, text)
			self.iter = self.editor.text_buffer.get_end_iter()
			handle.read(self.editor.opt_size, self.read_cb)

		except TypeError:

			pass

		return


	def _set_finalization_flags(self):
		"""
		Set flags at the end of loading a file.

		@param self: Reference to the FileLoader object.
		@type self: A FileLoader object.

		@note: See editor.py for comments on the flags.

		"""

		self.editor.loading_flag = False
		self.editor.dirty = True
		self.editor.isNewFile = False

		return


	def _get_file_information(self):
		"""
		Retrieve basic information about the file.

		The file size and optimal file size of the file are retrieved. The
		optimal file size is the best size chunk for loading files at a time.
		It differs among file systems and operating sytems. On most systems,
		the optimal file size is 4096. GNOMEVFS determines the optimal file
		size.

		@param self: Reference to the FileLoader object.
		@type self: A FileLoader object.

		"""

		file_info = get_file_info(self.editor.file_uri)
		self.editor.file_size = file_info.size
		self.editor.opt_size = file_info.io_block_size
		self.editor.opt_count = 0.0

		return


	def _show_error_dialog(self, message):
		"""
		Show an error dialog when a fatal event occurs.

		@param self: Reference to the FileLoader object.
		@type self: A FileLoader object.

		@param message: Message to display on the dialog window.
		@type message: A string object.

		"""

		show_cursor(self.editor)
		error_dialog(self.editor, message)
		self._error_recovery()

		return


	def launch_new_file_timeout(self, filename):
		"""
		Handles a timeout function called to launch a new editor after 2 seconds

		@param filename: name of a file
		@type filename: string

		@return: False to ensure the function is called only once
		@rtype:	boolean

		"""

		main.launch_new_file(filename, False)

		return False

